/*
  Program Name: xidx
  Copyright: Vercingetorix (vercingetorix11@gmail.com)
  Author: Vercingetorix (vercingetorix11@gmail.com)
  Date: 03/28/05
  License: This file is considered part of the program 'xidx'
*/


#ifndef _XIDX_MAIN_H
#define _XIDX_MAIN_H

#include <direct.h>
#include <string.h>
#include "shared.h"

#define HELP_HELP_STR "usage: help [command]\nexamples:\n\thelp chpath"

#define HELP_CHPATH_STR \
    "changes the filename paths in a skeleton file\n\n"\
    "usage: chpath <skeleton filename> [ALL] <filename> <new filename>\n"\
    "specifying \"ALL\" will match all strings as opposed to just the first\n"\ 
    "examples:\n\tchpath fs_new_skel ALL data/animations/LIS data/animations/NIS"\
    "\n\tchpath fs_new_skel \"data/animations/LIS 01 Stand Idle.cas\""\
    " \"data/animations/My new animation.cas\""
#define HELP_SKEL_STR "replaces the skeleton hierarchy with one from a .cas file\n" \
                      "usage: skel <skeleton filename> <cas filename>\n" \
                      "examples:\n\tskel fs_new_skel monster.cas"
#define HELP_SCALE_STR \
    "changes the scale of a skeleton\n" \
    "usage: scale <skeleton filename> [scale value]\n" \
    "if you do not specify a scale value it will print the current one"
    
#define HELP_CLEAR_STR "clears screen"

#define HELP_FORK_STR "passes the command to the system to be executed\n" \
                      "usage: fork <command>\n" \
                      "example:\n\tfork xidx -ts skeleton.idx"
    

/*globals...*/
extern bool  verbose;

//functions to go with shell
struct xidxFunc{    
    xidxFunc( const char* name, const char* usage, int (*execute)(void*) ){
        this->name  = name;
        this->usage = usage;
        this->execute = execute;
    }    
    const char* name;
    const char* usage;
    int (*execute)(void*);
};    

        
//XIDX 'shell'
class xidxShell{
    public:
        xidxShell();
        ~xidxShell();
    public:
        int Run();
        int Parse( char* );
    public:
        static int help( void* );
        static int chpath( void* );
        static int clear( void* );
        static int loadskel( void* );
        static int fork( void*);
        static int scale( void* );
    private:
        static xidxFunc* getfn( const char* );
        static stack<char*> gettokens( char*, const char*, bool=true );
        static int numargs( const char** );
    private:
        char input[256];
};    
    
#endif
